/*
** Module   : Pause.java
** Abstract : directive to pause for a fixed time interval in milliseconds
**
** Copyright (c) 2009-2023, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------------Description----------------------------------------
** 001 GES 20090114 Directive to pause for a fixed time interval in milliseconds.
** 002 CA  20210827 Refactored to allow other test types, beside terminal (which can be executed from the same 
**                  test suite).  Added SOAP and REST web service tests.
** 003 TJD 20220817 Implement performance test mode
** 004 TJD 20230119 Test timeout multiplier support
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You should have received a copy of the GNU Affero General Public License
** along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.goldencode.harness.test;

import com.goldencode.harness.*;
import com.goldencode.util.*;

/**
 * Directive to pause for a fixed time interval in milliseconds.
 */
public class Pause
extends Report
implements Testable
{
   /** Short name for the activity. */
   private static final String NAME = "PAUSE";
   
   /** Format specification for creating a description of this step. */
   private static final String SPEC = "millis = '%d';";
   
   /** Number of milliseconds to pause. */
   private int millis = 0;
   
   /** Runtime source for the milliseconds value to pause. */
   private Variable var = null;
   
   /**
    * Construct an instance of a step which pauses for a fixed time interval.
    *
    * @param    millis
    *           Number of milliseconds to pause.
    */
   public Pause(int millis)
   {
      super(NAME, String.format(SPEC, millis));
      this.millis = millis;
   }
   
   /**
    * Construct an instance of a step which pauses for a fixed time interval
    * as specified by the runtime value of the given variable.
    *
    * @param    var
    *           Source of the milliseconds to pause (read at the time of
    *           test execution, not now).
    */
   public Pause(Variable var)
   {
      super(NAME, "");
      this.var = var;
   }
   
   /**
    * Execute the directive of the test using the given terminal as the data
    * sink and/or source for the operation.
    */
   @Override
   public void execute()
   {
      Status status = Status.FAILED;
      String reason = null;
      
      if (var != null)
      {
         millis = var.asInt();
         setDescription(String.format(SPEC, millis));
      }
      
      try
      {
         if (Harness.isBenchmarkMode())
         {
            LogHelper.finer("Skipping pause:" + millis);
         }
         else
         {
            Thread.sleep(millis);
         }
         status = Status.PASSED;
      }
      
      catch (InterruptedException ie)
      {
         reason = String.format("Interrupted: %s", ie.getMessage());
      }
      
      setStatus(status, reason);
   }
   
   /**
    * Get the maximum number of milliseconds that this testable activity is allowed to run.
    * Valid for some of Tests only, -1 if not applicable
    * @return   The maximum number of milliseconds.
    */
   @Override
   public long getMaxElapsed()
   {
      if (millis > 0)
      {
         return Harness.scaleTimeoutMultiplier(millis);
      }
      return millis;
   }
}
